-- --------------------------------------------------------------------
-- **                              Data                              **
-- --------------------------------------------------------------------

EnhancedBars_Version = 8;

local whitelist = {
	["MainMenuExpBar"] = true,
	["TargetFrameHealthBar"] = true,
	["TargetFrameManaBar"] = true,
	["FocusFrameHealthBar"] = true,
	["FocusFrameManaBar"] = true,
};

local compactFrames = { };

local implementation = {
    {
        healthBar = "PlayerFrameHealthBar",
        raise = {
            { PlayerFrame:GetChildren(), 6 },
			{ PlayerFrameManaBar.FeedbackFrame, 7 },
			{ PlayerFrameManaBar.FullPowerFrame:GetChildren(), 7 },
        },
        unit = "player",
    },
    {
        healthBar = "PetFrameHealthBar",
        raise = {
            { select(2, PetFrame:GetChildren()), 6 },
        },
        unit = "pet",
    },
    {
        healthBar = "TargetFrameHealthBar",
        raise = {
            { TargetFrameToT, 6 },
            { TargetFrame:GetChildren(), 6 },
        },
        unit = "target",
    },
    {
        healthBar = "TargetFrameToTHealthBar",
        raise = {
            { TargetFrameToTTextureFrame, 6 },
        },
        unit = "targettarget",
    },
    {
        healthBar = "FocusFrameHealthBar",
        raise = {
            { FocusFrameToT, 6 },
            { FocusFrameTextureFrame, 6 },
        },
        unit = "focus",
    },
    {
        healthBar = "FocusFrameToTHealthBar",
        raise = {
            { FocusFrameToTTextureFrame, 6 },
        },
        unit = "focustarget",
    },
};

-- Add party frames
do
    for i=1, 4 do
        local frame = _G["PartyMemberFrame"..i];

        tinsert(implementation, {
            healthBar = frame:GetName().."HealthBar",
            raise = {
                { select(3, frame:GetChildren()), 5 },
            },
            unit = "party"..i,
        });
    end
end

-- Add arena frames
do
    -- This defeats the purpose of load-on-demand, but arena frames are quite light
    Arena_LoadUI();

    for i=1, 5 do
        local frame = _G["ArenaEnemyFrame"..i];

        tinsert(implementation, {
            healthBar = frame:GetName().."HealthBar",
            raise = {
                { select(2, frame:GetChildren()):GetChildren(), 5 },
            },
            unit = "arena"..i,
        });
    end
end

-- --------------------------------------------------------------------
-- **                            Functions                           **
-- --------------------------------------------------------------------

local function IsPriorityTarget(unit)
	return false;
--[[
	local guid = UnitGUID(unit);
	if not guid then return false; end
	local _, _, _, _, _, _, _, id, _ = guid:find("^(%w+)%-(%d+)%-(%d+)%-(%d+)%-(%d+)%-(%d+)%-(%x+)$");
	return id == "61245"; -- Lightning Surge Totem
]]
end

function EnhancedBars_InstallHealthBar(healthBar, isCompact)
    assert(healthBar);

	-- All of these seem broken in Legion (except absorb)
    healthBar.changeBar = EnhancedBars_CreateBar(healthBar, "CHANGE", isCompact);
    healthBar.absorbBar = EnhancedBars_CreateBar(healthBar, "ABSORB", isCompact);
    healthBar.necroBar = EnhancedBars_CreateBar(healthBar, "NECRO", isCompact);
    healthBar.overTimeBar = EnhancedBars_CreateBar(healthBar, "OVERTIME", isCompact);
	
	-- Remove Blizzard health loss animation
	if healthBar.AnimatedLossBar then
		healthBar.AnimatedLossBar:ClearAllPoints();
	end
end

function EnhancedBars_GetBalance(powerType, unit)
    if OverTimePredictor_GetBalance then
        local showHoT = EnhancedBars_Options["ShowHoT"];
        local showDoT = EnhancedBars_Options["ShowDoT"];
        local total, hot, dot = OverTimePredictor_GetBalance(powerType, unit);

        if showHoT and showDoT then
            return total;
        elseif showHoT then
            return hot;
        elseif showDoT then
            return -dot;
        end
    end
    return 0;
end

function EnhancedBars_GetAbsorb(unit)
    if OverTimePredictor_GetAbsorb then
        local showAbsorb = EnhancedBars_Options["ShowAbsorb"];
        local showNecro = EnhancedBars_Options["ShowNecro"];
        local total, absorb, necro = OverTimePredictor_GetAbsorb(unit);

        if not showAbsorb then absorb = 0; end
        if not showNecro then necro = 0; end
        return absorb, necro;
    end
    return 0, 0;
end

-- --------------------------------------------------------------------
-- **                               Hooks                            **
-- --------------------------------------------------------------------

function EnhancedBars_UpdateBar(bar, unit)
	if not unit then return; end
	
	local guid = UnitGUID(unit);
	if not guid then return; end
	
    local currentValue = UnitHealth(unit); -- bar:GetValue();
    local maxValue = UnitHealthMax(unit); -- select(2, bar:GetMinMaxValues());
    local absorb, necro = EnhancedBars_GetAbsorb(unit);
    local changeBar, absorbBar, necroBar, overTimeBar = bar.changeBar, bar.absorbBar, bar.necroBar, bar.overTimeBar;

    if changeBar then
        if EnhancedBars_Options["ShowChange"] then
            EnhancedBars_ModifyChangeBar(changeBar, currentValue, maxValue);
			
			-- Clear change bar if the unit changed
			EnhancedBars_VerifyGUID(changeBar, guid);
        end
    end

    if absorbBar then
        EnhancedBars_ModifyPredictedBar(absorbBar, currentValue + absorb, currentValue, maxValue);
        
        -- Hide Blizzard absorb if we display our own
		local frame = bar:GetParent();
        if EnhancedBars_Options["ShowAbsorb"] then
			-- Standard bar
            if frame.totalAbsorbBar then
                frame.totalAbsorbBar:Hide();
                frame.totalAbsorbBar.overlay:Hide();
                frame.overAbsorbGlow:Hide();
            end
            if frame.healAbsorbBar then
                frame.healAbsorbBar:Hide();
                frame.healAbsorbBarLeftShadow:Hide();
                frame.healAbsorbBarRightShadow:Hide();
                frame.overHealAbsorbGlow:Hide();
            end
			
			-- Compact bar
            if frame.totalAbsorb then
                frame.totalAbsorb:Hide();
                frame.totalAbsorbOverlay:Hide();
                frame.overAbsorbGlow:Hide();
            end
            if frame.myHealAbsorb then
                frame.myHealAbsorb:Hide();
                frame.myHealAbsorbLeftShadow:Hide();
                frame.myHealAbsorbRightShadow:Hide();
				frame.overHealAbsorbGlow:Hide();
            end
        end
    end

    if necroBar then
        EnhancedBars_ModifyPredictedBar(necroBar, currentValue - necro, currentValue, maxValue);
    end

    if overTimeBar then
        EnhancedBars_ModifyPredictedBar(overTimeBar, currentValue + EnhancedBars_GetBalance(-1, unit), currentValue, maxValue);
    end
end

function EnhancedBars_UnitFrameBar_Update(bar, unit)
    if unit ~= bar.unit then return; end
	if bar.disconnected or bar.lockValues then return; end
	
	EnhancedBars_UpdateBar(bar, unit);
end

function EnhancedBars_UnitFrameBar_OnUpdate(bar)
    EnhancedBars_UnitFrameBar_Update(bar, bar.unit);
end

function EnhancedBars_CompactUnitFrame_OnUpdate(frame)
	local name = frame:GetName();
	if name and compactFrames[name] then
		EnhancedBars_UpdateBar(frame.healthBar, frame.displayedUnit);
		
		if EnhancedBars_Options["HidePVPPetsHack"] then
			-- HACK: make non-player nameplates significantly transparent while in PVP combat
			-- Hopefully, we'll be able to remove it when Blizzard realizes that players don't care about pet nameplates in PVP
			if UnitIsPVP("player") and not UnitIsPlayer(frame.displayedUnit) and not IsPriorityTarget(frame.displayedUnit) then
				frame:SetAlpha(0.1);
			else
				frame:SetAlpha(1.0);
			end
		else
			frame:SetAlpha(1.0);
		end
	end
end

local function EnhancedBars_ClearText(statusFrame, textString)
	textString:SetText("");
	
	if statusFrame.LeftText and statusFrame.RightText then
		statusFrame.LeftText:SetText("");
		statusFrame.RightText:SetText("");
		statusFrame.LeftText:Hide();
		statusFrame.RightText:Hide();
	end
end

function EnhancedBars_TextStatusBar_UpdateTextStringWithValues(statusFrame, textString, value, valueMin, valueMax)
	if EnhancedBars_Options["ShowTextOnTargetAndFocusOnly"] and not whitelist[statusFrame:GetName()] then
		EnhancedBars_ClearText(statusFrame, textString);
		return;
	end
	
    if not EnhancedBars_Options["HideMaxText"] then
		return;
	end
	
    -- Alterations from Blizzard code starts here

	EnhancedBars_ClearText(statusFrame, textString);
	
	if ( ( tonumber(valueMax) ~= valueMax or valueMax > 0 ) and not ( statusFrame.pauseUpdates ) ) then
		statusFrame:Show();
		
		if ( (statusFrame.cvar and GetCVar(statusFrame.cvar) == "1" and statusFrame.textLockable) or statusFrame.forceShow ) then
			textString:Show();
		elseif ( statusFrame.lockShow > 0 and (not statusFrame.forceHideText) ) then
			textString:Show();
		else
			textString:SetText("");
			textString:Hide();
			return;
		end

		local valueDisplay = value;
		if ( statusFrame.capNumericDisplay ) then
			valueDisplay = AbbreviateLargeNumbers(value);
		else
			valueDisplay = BreakUpLargeNumbers(value);
		end

		local textDisplay = GetCVar("statusTextDisplay");
		if ( value and valueMax > 0 and ( textDisplay ~= "NUMERIC" or statusFrame.showPercentage ) and not statusFrame.showNumeric) then
			if ( value == 0 and statusFrame.zeroText ) then
				textString:SetText(statusFrame.zeroText);
				statusFrame.isZero = 1;
				textString:Show();
			elseif ( textDisplay == "BOTH" and not statusFrame.showPercentage) then
				if( statusFrame.LeftText and statusFrame.RightText ) then
					if(not statusFrame.powerToken or statusFrame.powerToken == "MANA") then
						statusFrame.LeftText:SetText(math.ceil((value / valueMax) * 100) .. "%");
						statusFrame.LeftText:Show();
					end
					statusFrame.RightText:SetText(valueDisplay);
					statusFrame.RightText:Show();
					textString:Hide();
				else
					valueDisplay = math.ceil((value / valueMax) * 100) .. "%   " .. valueDisplay;
				end
				textString:SetText(valueDisplay);
			else
				valueDisplay = math.ceil((value / valueMax) * 100) .. "%";
				if ( statusFrame.prefix and (statusFrame.alwaysPrefix or not (statusFrame.cvar and GetCVar(statusFrame.cvar) == "1" and statusFrame.textLockable) ) ) then
					textString:SetText(statusFrame.prefix .. "   " .. valueDisplay);
				else
					textString:SetText(valueDisplay);
				end
			end
		elseif ( value == 0 and statusFrame.zeroText ) then
			textString:SetText(statusFrame.zeroText);
			statusFrame.isZero = 1;
			textString:Show();
		else
			statusFrame.isZero = nil;
			if ( statusFrame.prefix and (statusFrame.alwaysPrefix or not (statusFrame.cvar and GetCVar(statusFrame.cvar) == "1" and statusFrame.textLockable) ) ) then
				textString:SetText(statusFrame.prefix.."   "..valueDisplay);
			else
				textString:SetText(valueDisplay);
			end
		end
	else
		textString:Hide();
		textString:SetText("");
		if ( not statusFrame.alwaysShow ) then
			statusFrame:Hide();
		else
			statusFrame:SetValue(0);
		end
	end
end

-- --------------------------------------------------------------------
-- **                             Handlers                           **
-- --------------------------------------------------------------------

function EnhancedBars_Install(info)
    EnhancedBars_InstallHealthBar(_G[info.healthBar], false);
	_G[info.healthBar]:HookScript("OnValueChanged", EnhancedBars_UnitFrameBar_OnUpdate); -- To be extra safe

    local raiseList = info.raise;
    for i=1, #raiseList do
        local info = raiseList[i];
        info[1]:SetFrameLevel(info[1]:GetFrameLevel() + info[2]);
    end
end

function EnhancedBars_InstallCompactUnitFrame(frame)
	local name = frame:GetName();
	if name then
		if name:sub(1, 9) == "NamePlate" and EnhancedBars_Options["IncludeNameplates"] then
			compactFrames[name] = true;
			EnhancedBars_InstallHealthBar(frame.healthBar, true);
		elseif name:sub(1, 7) == "Compact" and EnhancedBars_Options["IncludeRaidFrames"] then
			compactFrames[name] = true;
			EnhancedBars_InstallHealthBar(frame.healthBar, true);
		end
	end
end

do
    hooksecurefunc("UnitFrameHealthBar_Update", EnhancedBars_UnitFrameBar_Update);
    hooksecurefunc("UnitFrameHealthBar_OnUpdate", EnhancedBars_UnitFrameBar_OnUpdate);
	hooksecurefunc("UnitFrameHealPredictionBars_Update", EnhancedBars_UnitFrameBar_OnUpdate);
    hooksecurefunc("TextStatusBar_UpdateTextStringWithValues", EnhancedBars_TextStatusBar_UpdateTextStringWithValues);
	hooksecurefunc("CompactUnitFrame_OnLoad", EnhancedBars_InstallCompactUnitFrame);
	hooksecurefunc("CompactUnitFrame_OnUpdate", EnhancedBars_CompactUnitFrame_OnUpdate);
	hooksecurefunc("CompactUnitFrame_UpdateHealth", EnhancedBars_CompactUnitFrame_OnUpdate);
	hooksecurefunc("CompactUnitFrame_UpdateHealPrediction", EnhancedBars_CompactUnitFrame_OnUpdate);
		
    for index, info in pairs(implementation) do
        EnhancedBars_Install(info);
    end
end